<?php

/* vim: set expandtab shiftwidth=4 tabstop=4 softtabstop=4 foldmethod=marker: */

/**
 * Parser for MARC records
 * 
 * This package is based on the PHP MARC package, originally called "php-marc",
 * that is part of the Emilda Project (http://www.emilda.org). Christoffer
 * Landtman generously agreed to make the "php-marc" code available under the
 * GNU LGPL so it could be used as the basis of this PEAR package.
 * 
 * PHP version 5
 *
 * LICENSE: This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Christoffer Landtman <landtman@realnode.com>
 * @author     Dan Scott <dscott@laurentian.ca>
 * @copyright  2003-2006 Oy Realnode Ab, Dan Scott
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @version    CVS: $Id:$
 * @link       http://pear.php.net/package/File_MARC
 * @example    marc_yaz.php Pretty print a MARC record retrieved through the PECL yaz extension
 */

/**
 * To-dos:
 * 
 * @todo Ensure every function has an explicit return value
 *
 */

require_once 'PEAR/ErrorStack.php';
require_once 'Structures/Linked_List.php';

// {{{ constants

/**
 * MARC records retrieved from a file
 */
define("FILE_MARC_SOURCE_FILE", 1);

/**
 * MARC records retrieved from a binary string 
 */
define("FILE_MARC_SOURCE_STRING", 2);

/**
 * Hexadecimal value for Subfield indicator
 */
define("FILE_MARC_SUBFIELD_INDICATOR", "\x1F");

/**
 * Hexadecimal value for End of Field
 */
define("FILE_MARC_END_OF_FIELD", "\x1E");

/**
 * Hexadecimal value for End of Record
 */
define("FILE_MARC_END_OF_RECORD", "\x1D");

/**
 * Length of the Directory
 */
define("FILE_MARC_DIRECTORY_ENTRY_LEN", 12);

/**
 * Length of the Leader
 */
define("FILE_MARC_LEADER_LEN", 24);

/**
 * Add link to the end of the linked list
 */
define("FILE_MARC_ADD_APPEND", STRUCTURES_LINKED_LIST_ADD_APPEND);

/**
 * Add link before the target link in the linked list
 */
define("FILE_MARC_ADD_BEFORE", STRUCTURES_LINKED_LIST_ADD_BEFORE);

/**
 * Add link after the target link in the linked list
 */
define("FILE_MARC_ADD_AFTER", STRUCTURES_LINKED_LIST_ADD_AFTER);

/**
 * Define error codes
 */

/**
 * File could not be opened
 */
define("FILE_MARC_ERROR_INVALID_FILE", -1);

/**
 * User passed an unknown FILE_MARC_SOURCE_ constant
 */
define("FILE_MARC_ERROR_INVALID_SOURCE", -2);

/**
 * MARC record ended with an invalid terminator
 */
define("FILE_MARC_ERROR_INVALID_TERMINATOR", -3);

/**
 * No directory was found for the MARC record
 */
define("FILE_MARC_ERROR_NO_DIRECTORY", -4);

/**
 * An entry in the MARC directory was not 12 bytes
 */
define("FILE_MARC_ERROR_INVALID_DIRECTORY_LENGTH", -5);

/**
 * An entry in the MARC directory specified an invalid tag
 */
define("FILE_MARC_ERROR_INVALID_DIRECTORY_TAG", -6);

/**
 * An entry in the MARC directory specified an invalid tag length
 */
define("FILE_MARC_ERROR_INVALID_DIRECTORY_TAG_LENGTH", -7);

/**
 * An entry in the MARC directory specified an invalid field offset
 */
define("FILE_MARC_ERROR_INVALID_DIRECTORY_OFFSET", -8);

/**
 * An entry in the MARC directory runs past the end of the record
 */
define("FILE_MARC_ERROR_INVALID_DIRECTORY", -9);

/**
 * A field does not end with the expected end-of-field character
 */
define("FILE_MARC_ERROR_FIELD_EOF", -10);

/**
 * A field has invalid indicators
 */
define("FILE_MARC_ERROR_INVALID_INDICATORS", -11);

/**
 * A subfield is defined, but has no data
 */
define("FILE_MARC_ERROR_EMPTY_SUBFIELD", -12);

/**
 * An indicator other than 1 or 2 was requested
 */
define("FILE_MARC_ERROR_INVALID_INDICATOR_REQUEST", -13);

/**
 * An invalid mode for adding a field was specified
 */
define("FILE_MARC_ERROR_ADDFIELD_MODE", -14);

/**
 * An invalid object was passed instead of a File_MARC_Field object
 */
define("FILE_MARC_ERROR_INVALID_FIELD", -15);

/**
 * An invalid object was passed instead of a File_MARC_Subfield object
 */
define("FILE_MARC_ERROR_INVALID_SUBFIELD", -16);

/**
 * An invalid mode for adding a subfield was specified
 */
define("FILE_MARC_ERROR_ADDSUBFIELD_MODE", -17);

/**
 * The length in the MARC leader does not match the actual record length
 */
define("FILE_MARC_ERROR_INCORRECT_LENGTH", -18);

/**
 * The length field in the leader was less than five characters long
 */
define("FILE_MARC_ERROR_MISSING_LENGTH", -19);

/**
 * A five-digit length could not be found in the MARC leader
 */
define("FILE_MARC_ERROR_NONNUMERIC_LENGTH", -20);

/**
 * Tag does not adhere to MARC standards
 */
define("FILE_MARC_ERROR_INVALID_TAG", -21);

/**
 * A field has invalid indicators
 */
define("FILE_MARC_WARN_INVALID_INDICATOR", 1);
// }}}

//{{{ GLOBALS
/**
 * Error message stack
 * @global ErrorStack $GLOBALS["_File_MARC_stack"]
 */
$GLOBALS["_File_MARC_stack"] =& PEAR_ErrorStack::singleton('File_MARC');

/**
 * Error messages
 * @global array $GLOBALS["_File_MARC_messages"]
 */
$GLOBALS["_File_MARC_messages"] = array(
    FILE_MARC_ERROR_EMPTY_SUBFIELD => 'No subfield data found in tag "%tag%"', 
    FILE_MARC_ERROR_FIELD_EOF => 'Field for tag "%tag%" does not end with an end of field character', 
    FILE_MARC_ERROR_INCORRECT_LENGTH => 'Invalid record length: Leader says "%record_length%" bytes; actual record length is "%actual%"', 
    FILE_MARC_ERROR_ADDFIELD_MODE => 'addField() mode "%mode%" was not recognized', 
    FILE_MARC_ERROR_ADDSUBFIELD_MODE => 'addSubfield() mode "%mode%" was not recognized', 
    FILE_MARC_ERROR_INVALID_DIRECTORY => 'Directory entry for tag "%tag%" runs past the end of the record', 
    FILE_MARC_ERROR_INVALID_DIRECTORY_LENGTH => 'Invalid directory length', 
    FILE_MARC_ERROR_INVALID_DIRECTORY_OFFSET => 'Invalid offset "%offset%" for tag "%tag%" in directory', 
    FILE_MARC_ERROR_INVALID_DIRECTORY_TAG => 'Invalid tag "%tag%" in directory', 
    FILE_MARC_ERROR_INVALID_DIRECTORY_TAG_LENGTH => 'Invalid length "%len%" in directory for tag "%tag%"', 
    FILE_MARC_ERROR_INVALID_FIELD => 'Specified field must be a File_MARC_Data_Field or File_MARC_Control_Field object, but was "%field%"', 
    FILE_MARC_ERROR_INVALID_FILE => 'Invalid input file "%filename%"', 
    FILE_MARC_ERROR_INVALID_INDICATOR_REQUEST => 'Attempt to access indicator "%indicator%" failed; 1 and 2 are the only valid indicators', 
    FILE_MARC_ERROR_INVALID_INDICATORS => 'Invalid indicators "%indicators%" forced to blanks for tag "%tag%"', 
    FILE_MARC_ERROR_INVALID_SOURCE => "Invalid source for MARC records", 
    FILE_MARC_ERROR_INVALID_SUBFIELD => 'Specified field must be a File_MARC_Subfield object, but was "%class%"', 
    FILE_MARC_ERROR_INVALID_TAG => 'Tag "%tag%" is not a valid tag.', 
    FILE_MARC_ERROR_INVALID_TERMINATOR => 'Invalid record terminator', 
    FILE_MARC_ERROR_MISSING_LENGTH => "Couldn't find record length", 
    FILE_MARC_ERROR_NO_DIRECTORY => 'No directory found', 
    FILE_MARC_ERROR_NONNUMERIC_LENGTH => 'Record length "%record_length%" is not numeric',
    FILE_MARC_WARN_INVALID_INDICATOR => 'Illegal indicator "%indicator%" in field "%tag%" forced to blank', 
);
// }}}

//{{{ Error handling with PEAR::ErrorStack
$GLOBALS["_File_MARC_stack"]->setErrorMessageTemplate($GLOBALS["_File_MARC_messages"]);
//}}}

// {{{ class File_MARC
/**
 * The main File_MARC class enables you to read MARC records from a stream or string.
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Christoffer Landtman <landtman@realnode.com>
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/File_MARC
 */
class File_MARC {

    // {{{ properties
    /**
     * Array containing raw records
     * 
     * @var array
     */
    protected $raw;

    /**
     * Current position in the array of records
     *
     * @var integer
     */
    protected $pointer;
    // }}}

    // {{{ Constructor: function __construct()
    /**
     * Read in MARC records
     *
     * This function reads in MARC record files or strings that
     * contain one or more MARC records.
     *
     * <code>
     * <?php
     * // Retrieve MARC records from a file
     * $journals = new File_MARC('journals.mrc', FILE_MARC_SOURCE_FILE);
     *
     * // Retrieve MARC records from a string (e.g. Z39 query results)
     * $monographs = new File_MARC($raw_marc, FILE_MARC_SOURCE_STRING);
     * ?>
     * </code>
     *
     * @param string $in Name of the file, or a raw MARC string
     * @param int $source Source of the input, either FILE_MARC_SOURCE_FILE or FILE_MARC_SOURCE_STRING
     * @return bool Returns false if MARC records cannot be read
     */
    function __construct($source, $type = FILE_MARC_SOURCE_FILE)
    {
        switch ($type) {

        case FILE_MARC_SOURCE_FILE:
            if (!file_exists($source)) {
                $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_FILE, 'error', array('filename' => $source));
                return false;
            }
            $input = file($source);
            $recs  = explode(FILE_MARC_END_OF_RECORD, join("", $input));
            // Append FILE_MARC_END_OF_RECORD as we lost it when splitting
            // Last is not record, as it is empty because every record ends
            // with FILE_MARC_END_OF_RECORD.
            for ($i = 0; $i < (count($recs)-1); $i++) {
                $this->raw[] = $recs[$i].FILE_MARC_END_OF_RECORD;
            }
            $this->pointer = 0;
            return true;
            break;

        case FILE_MARC_SOURCE_STRING:
            $this->raw[] = $source;
            $this->pointer = 0;
            return true;
            break;

        default:
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_SOURCE, 'error');
            return false;
        }
    }
    // }}}

    // {{{ nextRaw()
    /**
     * Return the next raw MARC record
     *
     * Returns the next raw MARC record, unless all
     * records already have been read.
     *
     * @return string Either a raw record or false
     */
    function nextRaw()
    {
        /**
         * Exit if we are at the end of the file
         */
        if ($this->pointer >= count($this->raw)) {
            return false;
        }

        /**
         * Read next line
         */
        $record = $this->raw[$this->pointer++];

        // remove illegal stuff that sometimes occurs between records
        $record = preg_replace('/^[\\x0a\\x0d\\x00]+/', "", $record);

        /**
         * Record validation
         */
        if (strlen($record) < 5) {
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_MISSING_LENGTH, 'error');
        }
        $record_length = substr($record, 0, 5);
        if (!preg_match("/^\d{5}$/", $record_length)) {
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_NONNUMERIC_LENGTH, 'error', array("record_length" => $record_length));
        }
        if ($record_length != strlen($record)) {
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INCORRECT_LENGTH, 'error', array("record_length" => $record_length, "actual" => strlen($record)));
        }

        return $record;
    }
    // }}}

    // {{{ next()
    /**
     * Return next File_MARC_Record object
     *
     * Decode the next raw MARC record and return
     * @return File_MARC_Record A File_MARC_Record object, or false if there are no more records
     */
    function next()
    {
        if ($raw = $this->nextRaw()) {
            return $this->_decode($raw);
        } else {
            return false;
        }
    }
    // }}}

    // {{{ _decode()
    /**
     * Decode a given raw MARC record
     *
     * "Port" of Andy Lesters MARC::File::USMARC->decode() Perl function into PHP.
     *
     * @param string $text Raw MARC record
     * @return File_MARC_Record Decoded File_MARC_Record object
     */
    private function _decode($text)
    {
        if (!preg_match("/^(\d{5})/", $text, $matches)) {
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_NONNUMERIC_LENGTH, 'error', array("record_length" => substr($text, 0, 5)));
        }

        $marc = new File_MARC_Record();

        // Store record length
        $record_length = $matches[1];

        if ($record_length != strlen($text)) {
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INCORRECT_LENGTH, 'error', array("record_length" => substr($record_length), "actual" => strlen($text)));
        }

        if (substr($text, -1, 1) != FILE_MARC_END_OF_RECORD)
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_TERMINATOR, 'error');

        // Store leader
        $marc->setLeader(substr($text, 0, FILE_MARC_LEADER_LEN));

        // bytes 12 - 16 of leader give offset to the body of the record
        $data_start = 0 + substr($text, 12, 5);

        // immediately after the leader comes the directory (no separator)
        $dir = substr($text, FILE_MARC_LEADER_LEN, $data_start - FILE_MARC_LEADER_LEN - 1);  // -1 to allow for \x1e at end of directory

        // character after the directory must be \x1e
        if (substr($text, $data_start-1, 1) != FILE_MARC_END_OF_FIELD) {
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_NO_DIRECTORY, 'error');
        }

        // All directory entries 12 bytes long, so length % 12 must be 0
        if (strlen($dir) % FILE_MARC_DIRECTORY_ENTRY_LEN != 0) {
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_DIRECTORY_LENGTH, 'error');
        }

        // go through all the fields
        $nfields = strlen($dir) / FILE_MARC_DIRECTORY_ENTRY_LEN;
        for ($n=0; $n<$nfields; $n++) {
            // As pack returns to key 1, leave place 0 in list empty
            list(, $tag) = unpack("A3", substr($dir, $n*FILE_MARC_DIRECTORY_ENTRY_LEN, FILE_MARC_DIRECTORY_ENTRY_LEN));
            list(, $len) = unpack("A3/A4", substr($dir, $n*FILE_MARC_DIRECTORY_ENTRY_LEN, FILE_MARC_DIRECTORY_ENTRY_LEN));
            list(, $offset) = unpack("A3/A4/A5", substr($dir, $n*FILE_MARC_DIRECTORY_ENTRY_LEN, FILE_MARC_DIRECTORY_ENTRY_LEN));

            // Check directory validity
            if (!preg_match("/^[0-9A-Za-z]{3}$/", $tag)) {
                $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_DIRECTORY_TAG, 'error', array("tag" => $tag));
            }
            if (!preg_match("/^\d{4}$/", $len)) {
                $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_DIRECTORY_TAG_LENGTH, 'error', array("tag" => $tag, "len" => $len));
            }
            if (!preg_match("/^\d{5}$/", $offset)) {
                $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_DIRECTORY_OFFSET, 'error', array("tag" => $tag, "offset" => $offset));
            }
            if ($offset + $len > $record_length) {
                $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_DIRECTORY, 'error', array("tag" => $tag));
            }

            $tag_data = substr($text, $data_start + $offset, $len);

            if (substr($tag_data, -1, 1) == FILE_MARC_END_OF_FIELD) {
                /* get rid of the end-of-tag character */
                $tag_data = substr($tag_data, 0, -1);
                $len--;
            } else {
                $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_FIELD_EOF, 'error', array("tag" => $tag));
            }

            if (preg_match("/^\d+$/", $tag) and ($tag < 10)) {
                $marc->addField(new File_MARC_Control_Field($tag, $tag_data));
            } else {
                $subfields = split(FILE_MARC_SUBFIELD_INDICATOR, $tag_data);
                $indicators = array_shift($subfields);

                if (strlen($indicators) > 2 || strlen($indicators) == 0) {
                    $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_INDICATORS, 'error', array("tag" => $tag, "indicators" => $indicators));
                    list($ind1,$ind2) = array(" ", " ");
                } else {
                    $ind1 = substr($indicators, 0, 1);
                    $ind2 = substr($indicators, 1, 1);
                }

                // Split the subfield data into subfield name and data pairs
                $subfield_data = array();
                foreach ($subfields as $subfield) {
                    if (strlen($subfield) > 0) {
                        $subfield_data[] = new File_MARC_Subfield(substr($subfield, 0, 1), substr($subfield, 1));
                    } else {
                        $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_EMPTY_SUBFIELD, 'error', array("tag" => $tag));
                    }
                }

                if (!isset($subfield_data)) {
                    $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_EMPTY_SUBFIELD, 'error', array("tag" => $tag));
                }

                $marc->addField(new File_MARC_Data_Field($tag, $subfield_data, $ind1, $ind2));
            }
        }

       return $marc;
    }
    // }}}

    // {{{ numRecords()
    /**
     * Get the number of records available in this Record
     * @return int The number of records
     */
    function numRecords()
    {
        return count($this->raw);
    }
    // }}}

    // {{{ getErrors()
    /**
     * Get errors (if any) for this object 
     *
     * Returns the error stack.
     *
     * @return PEAR_ErrorStack returns the error stack
     */
    function getErrors()
    {
        return $GLOBALS["_File_MARC_stack"];
    }
    // }}}
}
// }}}

// {{{ class File_MARC_Record
/**
 * Represents a single MARC record
 * 
 * A MARC record contains a leader and zero or more fields held within a
 * linked list structure. Fields are represented by {@link File_MARC_Data_Field} objects.
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Christoffer Landtman <landtman@realnode.com>
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/File_MARC
 */
class File_MARC_Record {

    // {{{ properties
    /**
     * Contains a linked list of {@link File_MARC_Data_Field} objects for this Record
     * @var Structures_Linked_List
     */
    protected $fields;

    /**
     * Record leader
     * @var string
     */
    protected $leader;

    // }}}

    // {{{ Constructor: function __construct()
    /**
     * Start function
     *
     * Set all variables to defaults to create new File_MARC_Record object
     * @return true 
     */
    function __construct()
    {
        $this->fields = new Structures_Linked_List();
        $this->leader = str_repeat(' ', 24);
    }
    // }}}

    // {{{ getErrors()
    /**
     * Get errors (if any) for this object 
     *
     * Returns the error stack.
     *
     * @return PEAR_ErrorStack returns the error stack
     */
    function getErrors()
    {
        return $GLOBALS["_File_MARC_stack"];
    }
    // }}}

    // {{{ getLeader()
    /**
     * Get MARC leader
     *
     * Returns the leader for the MARC record. No validation
     * on the specified leader is performed
     *
     * @return string returns the leader
     */
    function getLeader()
    {
        return $this->leader;
    }
    // }}}

    // {{{ setLeader()
    /**
     * Set MARC record leader
     *
     * Sets the leader for the MARC record. No validation
     * on the specified leader is performed.
     *
     * @param string $leader Leader
     * @return string returns the leader
     */
    function setLeader($leader)
    {
        $this->leader = $leader;
        return $this->leader;
    }
    // }}}

    // {{{ addField()
    /**
     * Adds field to MARC record
     *
     * Adds a File_MARC_Control_Field or File_MARC_Data_Field object to the existing list
     * of fields according to the specified mode.
     *
     * @param File_MARC_Field $new_field The field to add
     * @param int $mode specifies how to add the field
     * @param File_MARC_Field $existing_field specifies the reference field for locating the new field
     * @return File_MARC_Field the new File_MARC_Field object if the operation succeeded, or
     * a PEAR_Error object if the operation failed
     */
    function addField($new_field, $mode=FILE_MARC_ADD_APPEND, $existing_field=null)
    {
        if (!($new_field instanceof File_MARC_Field)) {
            if ($existing_field and !($existing_field instanceof File_MARC_Data_Field)) {
                $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_FIELD, 'error', array("field" => get_class($existing_field)));
            }
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_FIELD, 'error', array("field" => get_class($new_field)));
            return false;
        }

        switch ($mode) {
        /* Append as the last field in the record */
        case FILE_MARC_ADD_APPEND:
            $this->fields->addNode($new_field, FILE_MARC_ADD_APPEND);
            break;

        /* Insert before the specified field in the record */
        case FILE_MARC_ADD_BEFORE:
            $this->fields->addNode($new_field, FILE_MARC_ADD_BEFORE, $existing_field);
            break;

        /* Insert after the specified field in the record */
        case FILE_MARC_ADD_AFTER:
            $this->fields->addNode($new_field, FILE_MARC_ADD_AFTER, $existing_field);
            break;

        default: 
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_ADDFIELD_MODE, 'error', array("mode" => $mode));
            return false;
        }
        return $new_field;
    }
    // }}}

    // {{{ _buildDirectory()
    /**
     * Build Record Directory
     *
     * Generate the directory of the Record according to existing data.
     * @return array Array ($fields, $directory, $total, $base_address)
     */
    private function _buildDirectory()
    {
        // Vars
        $fields = array();
        $directory = array();

        $data_end = 0;
        // Not sure we can iterate over a Structures_Linked_List
        foreach ($this->fields as $field) {
            // No empty fields allowed
            if (!$field->isEmpty()) {
                // Get data in raw format
                $str = $field->toRaw();
                $fields[] = $str;

                // Create directory entry
                $len = strlen($str);
                $direntry = sprintf("%03s%04d%05d", $field->getTag(), $len, $data_end);
                $directory[] = $direntry;
                $data_end += $len;
            }
        }

        /**
         * Rules from MARC::Record::USMARC
         */
        $base_address =
                FILE_MARC_LEADER_LEN +    // better be 24
                (count($directory) * FILE_MARC_DIRECTORY_ENTRY_LEN) +
                                // all the directory entries
                1;              // end-of-field marker


        $total =
                $base_address +  // stuff before first field
                $data_end +      // Length of the fields
                1;              // End-of-record marker



        return array($fields, $directory, $total, $base_address);
    }
    // }}}

    // {{{ setLeaderLengths()
    /**
     * Set MARC record leader lengths
     *
     * Set the Leader lengths of the record according to defaults specified in
     * {@link http://www.loc.gov/marc/bibliographic/ecbdldrd.html}
     *
     * @param int $record_length Record length
     * @param int $base_address Base address of data
     * @return bool Success or failure
     */
    function setLeaderLengths($record_length, $base_address)
    {
        if (!is_int($record_length)) {
            return false;
        }
        if (!is_int($base_address)) {
            return false;
        }

        // Set record length
        $this->leader = substr_replace($this->leader, sprintf("%05d", $record_length), 0, 5);
        $this->leader = substr_replace($this->leader, sprintf("%05d", $base_address), 12, 5);
        $this->leader = substr_replace($this->leader, '22', 10, 2);
        $this->leader = substr_replace($this->leader, '4500', 20, 4);
        return true;
    }
    // }}}

    // {{{ getFields()
    /**
     * Return an array containing all {@link File_MARC_Data_Field} or
     * {@link File_MARC_Control_Field} objects that match the specified tag name.
     * If the tag name is omitted all fields will be returned.
     *
     * @param string $spec tag name
     * @param bool $pcre if true, then match as a regular expression
     * @return array File_MARC_Data_Field or File_MARC_Control_Field objects that match the requested tag name
     */
    function getFields($spec = null, $pcre = false)
    {
        if ($spec) {
            $matches = array();
            foreach ($this->fields as $field) {
                if (($pcre
                     && preg_match("/$spec/", $field->getTag()))
                   || (!$pcre
                     && $spec == $field->getTag())
                ) {
                    $matches[] = $field;
                }
            }
            return $matches;
        } else {
            return $this->fields;
        }
    }
    // }}}

    // {{{ deleteFields()
    /**
     * Delete all occurrences of a field matching a tag name from the record.
     *
     * @param string $tag tag for the fields to be deleted
     * @param bool $pcre if true, then match as a regular expression
     * @return int number of fields that were deleted
     */
    function deleteFields($tag, $pcre = false)
    {
        $cnt = 0;
        foreach ($this->getFields() as $field) {
            if (($pcre
                 && preg_match("/$tag/", $field->getTag()))
               || (!$pcre
                 && $tag == $field->getTag())
            ) {
                $field->delete();
                // dodgy things happen when you delete a node from the middle of a linked list in the middle of iterating through that list
                $this->getFields()->rewind();
                $cnt++;
            }
        }
        return $cnt;
    }
    // }}}

    // {{{ duplicate()
    /**
     * Duplicate a record with all its fields and subfields
     *
     * @return File_MARC_Record Return the duplicated record
     */
    function duplicate()
    {
        $duplicate = serialize($this); 
        return unserialize($duplicate);
    }
    // }}}

    // {{{ output methods
    /**
     * ========== OUTPUT METHODS ==========
     */

    // {{{ toRaw()
    /**
     * Return the Record in raw MARC format.
     *
     * @return string Raw MARC data
     */
    function toRaw()
    {
        list($fields, $directory, $record_length, $base_address) = $this->_buildDirectory();
        $this->setLeaderLengths($record_length, $base_address);

        /**
         * Glue together all parts
         */
        return $this->leader.implode("", $directory).FILE_MARC_END_OF_FIELD.implode("", $fields).FILE_MARC_END_OF_RECORD;
    }
    // }}}

    // {{{ __toString()
    /**
     * Return the MARC record in a pretty printed string
     *
     * Return the MARC record in a pretty printed string.
     *
     * @return string Formatted representation of MARC record
     */
    function __toString()
    {
        // Begin output
        $formatted = "LDR ".$this->leader."\n";
        foreach ($this->fields as $field) {
            if (!$field->isEmpty()) {
                $formatted .= $field->__toString()."\n";
            }
        }
        return $formatted;
    }
    // }}}

    // }}}

}
// }}}

// {{{ class File_MARC_Field
/**
 * The File_MARC_Field class is expected to be extended to reflect the requirements of control and data fields.
 *
 * A MARC field consists of a tag name.
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Christoffer Landtman <landtman@realnode.com>
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/File_MARC
 */
class File_MARC_Field extends Structures_Linked_List_Node {

    // {{{ properties
    /**
     * The tag name of the Field
     * @var string
     */
    protected $tag;

    // }}}

    // {{{ Constructor: function __construct()
    /**
     * File_MARC_Field constructor
     *
     * Create a new {@link File_MARC_Field} object from passed arguments. We have to
     * define placeholders for the arguments required by child classes.
     *
     * @param string $tag tag
     * @param string $subfields placeholder for subfields or control data
     * @param string $ind1 placeholder for first indicator
     * @param string $ind2 placeholder for second indicator
     * @return string Returns warnings, if any, issued during parse
     */
    function __construct($tag, $subfields=null, $ind1=null, $ind2=null) 
    {
        $this->tag = $tag;

        // Check if valid tag
        if (!preg_match("/^[0-9A-Za-z]{3}$/", $tag)) {
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_TAG, 'error', array("tag" => $tag));
        }

    }
    // }}}

    // {{{ Destructor: function __destruct()
    /**
     * Destroys the data field
     */
    function __destruct()
    {
        $GLOBALS["_File_MARC_stack"] = null;
        $this->tag = null;
        parent::__destruct();
    }
    // }}}

    // {{{ getTag()
    /**
     * Returns the tag for this {@link File_MARC_Field} object
     *
     * @return string returns the tag number of the field
     */
    function getTag()
    {
        return $this->tag;
    }
    // }}}

    // {{{ setTag()
    /**
     * Sets the tag for this {@link File_MARC_Field} object
     *
     * @param string $tag new value for the tag
     * @return string returns the tag number of the field
     */
    function setTag($tag)
    {
        $this->tag = $tag;
        return $this->getTag();
    }
    // }}}

    // {{{ isEmpty()
    /**
     * Is empty
     *
     * Checks if the field is empty.
     *
     * @return bool Returns true if the field is empty, otherwise false
     */
    function isEmpty()
    {
        if ($this->getTag()) {
            return false;
        }
        // It is empty
        return true;
    }
    // }}}

    // {{{ getErrors()
    /**
     * Get errors (if any) for this object 
     *
     * Returns the error stack.
     *
     * @return PEAR_ErrorStack returns the error stack
     */
    function getErrors()
    {
        return $GLOBALS["_File_MARC_stack"];
    }
    // }}}

    /**
     * ========== OUTPUT METHODS ==========
     */

    // {{{ __toString()
    /**
     * Return Field formatted
     *
     * Return Field as a formatted string.
     *
     * @return string Formatted output of Field
     */
    function __toString()
    {
        return $this->getTag();
    }
    // }}}

    // {{{ toRaw()
    /**
     * Return field in raw MARC format (stub)
     *
     * Return the field formatted in raw MARC for saving into MARC files. This
     * stub method will be extended by the child classes.
     *
     * @return bool Raw MARC
     */
    function toRaw()
    {
        return false;
    }
    // }}}
}
// }}}

// {{{ class File_MARC_Data_Field
/**
 * The File_MARC_Data_Field class represents a single field in a MARC record.
 *
 * A MARC field consists of a tag name, two indicators which may be null,
 * and zero or more subfields represented by {@link File_MARC_Subfield} objects.
 * Subfields are held within a linked list structure.
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Christoffer Landtman <landtman@realnode.com>
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/File_MARC
 */
class File_MARC_Data_Field extends File_MARC_Field {

    // {{{ properties
    /**
     * Value of the first indicator
     * @var string
     */
    protected $ind1;

    /**
     * Value of the second indicator
     * @var string
     */
    protected $ind2;

    /**
     * Linked list of subfields
     * @var Structures_Linked_List
     */
    protected $subfields;

    // }}}

    // {{{ Constructor: function __construct()
    /**
     * Field init function
     *
     * Create a new {@link File_MARC_Data_Field} object from passed arguments
     *
     * @param string $tag tag
     * @param array $subfields array of {@link File_MARC_Subfield} objects
     * @param string $ind1 first indicator
     * @param string $ind2 second indicator
     * @return string Returns warnings, if any, issued during parse
     */
    function __construct($tag, $subfields, $ind1=null, $ind2=null) 
    {
        $this->subfields = new Structures_Linked_List();

        parent::__construct($tag);

        $this->ind1 = $this->_validateIndicator($ind1);
        $this->ind2 = $this->_validateIndicator($ind2);

        // we'll let users add subfields after if they so desire
        if ($subfields) {
            $this->addSubfields($subfields);
        }
    }
    // }}}

    // {{{ Destructor: function __destruct()
    /**
     * Destroys the data field
     */
    function __destruct()
    {
        $this->subfields = null;
        $this->ind1 = null;
        $this->ind2 = null;
        parent::__destruct();
    }
    // }}}

    // {{{ Explicit destructor: function delete()
    /**
     * Destroys the data field
     */
    function delete()
    {
        $this->__destruct();
    }
    // }}}

    // {{{ protected function _validateIndicator()
    /**
     * Validate indicator field
     *
     * Validates the value passed in for an indicator
     *
     * @param string $indicator Value of the indicator to be validated
     * @return string Returns warnings, if any, issued during parse
     */
    private function _validateIndicator($indicator)
    {
        if (!preg_match("/^[0-9A-Za-z ]$/", $indicator)) {
            if ($indicator != "") {
                $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_WARN_INVALID_INDICATOR, 'warning', array("tag" => $this->getTag(), "indicator" => $indicator));
            }
            $indicator = " ";
        }
        return $indicator;
    }
    // }}}

    // {{{ addSubfield()
    /**
     * Add subfield to a field in a MARC record
     *
     * Given {@link File_MARC_Subfield} object will be added to the existing list of fields.
     * @param File_MARC_Subfield $newSubfield subfield to add
     * @param int $mode where to add the new subfield
     * @param File_MARC_Subfield $existingSubfield reference location for the new subfield
     * @return File_MARC_Subfield the new {@link File_MARC_Subfield} object if the operation succeeded, or
     * a PEAR_Error object if the operation failed
     */
    function addSubfield($newSubfield, $mode=FILE_MARC_ADD_APPEND, $existingSubfield=null)
    {
        if (!($newSubfield instanceof File_MARC_Subfield)) {
            if ($existingSubfield != null and !($existingSubfield instanceof File_MARC_Subfield)) {
                $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_SUBFIELD, 'error', array('class' => get_class($existingSubfield)));
            }
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_SUBFIELD, 'error', array('class' => get_class($newSubfield)));
            return false;
        }

        switch ($mode) {
        case FILE_MARC_ADD_APPEND:
            $this->subfields->addNode($newSubfield, FILE_MARC_ADD_APPEND);
            break;

        case FILE_MARC_ADD_BEFORE:
            $this->subfields->addNode($newSubfield, FILE_MARC_ADD_BEFORE, $existingSubfield);
            break;

        case FILE_MARC_ADD_AFTER:
            $this->subfields->addNode($newSubfield, FILE_MARC_ADD_AFTER, $existingSubfield);
            break;

        default:
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_ADDSUBFIELD_MODE, 'error', array("mode" => $mode));
            return false;
        }
        return $newSubfield;
    }
    // }}}

    // {{{ addSubfields()
    /**
     * Adds an array of subfields to a {@link File_MARC_Data_Field} object
     *
     * Appends subfields to existing subfields in the order in which
     * they appear in the array. For finer grained control of the subfield
     * order, use {@link addSubfield()}.
     *
     * @param array $subfields array of {@link File_MARC_Subfield} objects
     * @return int returns the number of subfields that were added
     */
    function addSubfields($subfields)
    {
        // Just in case someone passes in a single File_MARC_Subfield instead of an array
        if ($subfields instanceof File_MARC_Subfield) {
            $this->addSubfield($subfields);
            return 1;
        }

        // Add subfields
        $cnt = 0;
        foreach ($subfields as $subfield) {
            $this->addSubfield($subfield);
            $cnt++;
        }

        return $cnt;
    }
    // }}}

    // {{{ deleteSubfield()
    function deleteSubfield($subfield)
    {
        $this->subfields->deleteNode($subfield);
    }
    // }}}

    // {{{ getIndicator()
    /**
     * Get values of indicators
     *
     * @param int $ind number of the indicator (1 or 2)
     * @return string returns indicator value if it exists, otherwise false
     */
    function getIndicator($ind)
    {
        if ($ind == 1) {
            return $this->ind1;
        } elseif ($ind == 2) {
            return $this->ind2;
        } else {
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_INDICATOR_REQUEST, 'error', array("indicator" => $indicator));
        }
        return false;
    }
    // }}}

    // {{{ setIndicator()
    /**
     * Set value of an indicator
     *
     * @param int $ind number of the indicator (1 or 2)
     * @param string $value value of the indicator
     * @return string returns indicator value if it exists, otherwise false
     */
    function setIndicator($ind, $value)
    {
        switch ($ind) {

        case 1:
            $this->ind1 = $this->_validateIndicator($value);
            break;

        case 2:
            $this->ind2 = $this->_validateIndicator($value);
            break;

        default:
            $GLOBALS["_File_MARC_stack"]->push(FILE_MARC_ERROR_INVALID_INDICATOR_REQUEST, 'error', array("indicator" => $ind));
            return false;
        }

        return getIndicator($ind);
    }
    // }}}

    // {{{ &getSubfield()
    /**
     * Returns the {@link File_MARC_Subfield} object, or array of objects, for a requested code
     *
     * @param string $code subfield code for which the {@link File_MARC_Subfield} is retrieved
     * @return File_MARC_Subfield|array returns a single {@link File_MARC_Subfield} object, or an
     * array of {@link File_MARC_Subfield} objects, or false if no codes match $code
     */
    function &getSubfield($code)
    {
        $results = array();
        // iterate merrily through the subfields looking for the requested code
        $sf = $this->subfields->rewind();
        do {
            if (!$sf) {
                // there are no subfields at all; get out of here
                break;
            }
            if ($sf->getCode() == $code) {
                $results[] = $sf;
            }
        } while ($sf =& $this->subfields->next());

        switch (count($results)) {
            case 0:
                $results = false;
                return $results;
                break;

            case 1:
                return $results[0];
                break;

            default:
                return $results;
                break;
        }
    }
    // }}}

    // {{{ getAllSubfields()
    /**
     * Returns an array containing all subfields in this field
     *
     * @return Structures_Linked_List returns a {@link Structure_Linked_List} linked list of {@link File_MARC_Subfield} objects
     */
    function getAllSubfields()
    {
        return $this->subfields;
    }
    // }}}

    // {{{ isEmpty()
    /**
     * Is empty
     *
     * Checks if the field is empty. Criteria is that it has
     * at least one subfield with data
     * @return bool Returns true if the field is empty, otherwise false
     */
    function isEmpty()
    {
        foreach ($this->subfields as $subfield) {
            // Check if subfield has data
            if (!$subfield->isEmpty()) {
                return false;
            }
        }
        // It is empty
        return true;
    }
    // }}}

    /**
     * ========== OUTPUT METHODS ==========
     */

    // {{{ __toString()
    /**
     * Return Field formatted
     *
     * Return Field as a formatted string.
     *
     * @return string Formatted output of Field
     */
    function __toString()
    {
        // Variables
        $lines = array();
        // Process tag and indicators
        $pre = sprintf("%3s %1s%1s", $this->tag, $this->ind1, $this->ind2);

        // Process subfields
        $this->subfields->rewind();
        foreach ($this->subfields as $subfield) {
            $lines[] = sprintf("%6s _%1s%s", $pre, $subfield->getCode(), $subfield->getData());
            $pre = "";
        }

        return join("\n", $lines);
    }
    // }}}

    // {{{ toRaw()
    /**
     * Return Field in Raw MARC
     *
     * Return the Field formatted in Raw MARC for saving into MARC files
     *
     * @return string Raw MARC
     */
    function toRaw()
    {
        $subfields = array();
        foreach ($this->subfields as $subfield) {
            if (!$subfield->isEmpty()) {
                $subfields[] = $subfield->toRaw();
            }
        }
        return $this->ind1.$this->ind2.implode("", $subfields).FILE_MARC_END_OF_FIELD;
    }
    // }}}
}
// }}}

// {{{ class File_MARC_Control_Field
/**
 * The File_MARC_Control_Field class represents a single control field in a MARC record.
 *
 * A MARC control field consists of a tag name and control data.
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Christoffer Landtman <landtman@realnode.com>
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/File_MARC
 */
class File_MARC_Control_Field extends File_MARC_Field {

    // {{{ Properties
    /**
     * Value of field, if field is a Control field
     * @var string
     */
    protected $data;
    // }}}

    // {{{ Constructor: function __construct()
    /**
     * Field init function
     *
     * Create a new {@link File_MARC_Control_Field} object from passed arguments
     *
     * @param string $tag tag
     * @param string $data control field data
     * @param string $ind1 placeholder for class strictness
     * @param string $ind2 placeholder for class strictness
     * @return string Returns warnings, if any, issued during parse
     */
    function __construct($tag, $data, $ind1=null, $ind2=null) 
    {
        $this->data = $data;
        parent::__construct($tag);

    }
    // }}}

    // {{{ Destructor: function __destruct()
    /**
     * Destroys the control field
     */
    function __destruct()
    {
        $this->data = null;
        parent::__destruct();
    }
    // }}}

    // {{{ Explicit destructor: function delete()
    /**
     * Destroys the control field
     */
    function delete()
    {
        $this->__destruct();
    }
    // }}}

    // {{{ getData()
    /**
     * Get control field data
     *
     * @return string returns data in control field
     */
    function getData()
    {
        return $this->data;
    }
    // }}}

    // {{{ isEmpty()
    /**
     * Is empty
     *
     * Checks if the field contains data
     * @return bool Returns true if the field is empty, otherwise false
     */
    function isEmpty()
    {
        return ($this->data) ? false : true;
    }
    // }}}

    // {{{ setData()
    /**
     * Set control field data
     *
     * @param string $data data for the control field
     * @return bool returns the new data in the control field 
     */
    function setData($data)
    {
        $this->data = $data;
        return $this->getData();
    }
    // }}}

    // {{{ __toString()
    /**
     * Return as a formatted string
     *
     * Return the control field as a formatted string for pretty printing
     *
     * @return string Formatted output of control Field
     */
    function __toString()
    {
        return sprintf("%3s     %s", $this->tag, $this->data);
    }
    // }}}

    // {{{ toRaw()
    /**
     * Return as raw MARC
     *
     * Return the control field formatted in Raw MARC for saving into MARC files
     *
     * @return string Raw MARC
     */
    function toRaw()
    {
        return $this->data.FILE_MARC_END_OF_FIELD;
    }
    // }}}

}
// }}}

// {{{ class File_MARC_Subfield
/**
 * The File_MARC_Subfield class represents a single subfield in a MARC record field.
 *
 * Represents a subfield within a MARC field. This class also implements the possibility of
 * duplicate subfields within a single field, i.e. 650 _z Test1 _z Test2 for example. This class
 * also implements all management functions related to a single subfield.
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Christoffer Landtman <landtman@realnode.com>
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/File_MARC
 */
class File_MARC_Subfield extends Structures_Linked_List_Node {
    // {{{ properties
    /**
     * Subfield code, e.g. _a, _b
     * @var string
     */
    protected $code;

    /**
     * Data contained by the subfield
     * @var string
     */
    protected $data;
    // }}}

    // {{{ Constructor: function __construct()
    /**
     * File_MARC_Subfield constructor
     *
     * Create a new subfield to represent the code and data
     *
     * @param string $code Subfield code
     * @param string $data Subfield data
     */
    function __construct($code, $data)
    {
        $this->code = $code;
        $this->data = $data;
    }
    // }}}

    // {{{ Destructor: function __destruct()
    /**
     * Destroys the subfield
     */
    function __destruct()
    {
        $GLOBALS["_File_MARC_stack"] = null;
        $this->code = null;
        $this->data = null;
        parent::__destruct();
    }
    // }}}

    // {{{ Explicit destructor: function delete()
    /**
     * Destroys the subfield
     */
    function delete()
    {
        $this->__destruct();
    }
    // }}}

    // {{{ getCode()
    /**
     * Return code of the subfield
     *
     * @return string Tag name
     */
    function getCode()
    {
        return $this->code;
    }
    // }}}

    // {{{ getData()
    /**
     * Return data of the subfield
     *
     * @return string data
     */
    function getData()
    {
        return $this->data;
    }
    // }}}

    // {{{ __toString()
    /**
     * Return string representation of subfield
     *
     * @return string String representation
     */
    public function __toString()
    {
        $pretty = '[' . $this->getCode() . ']: ' . $this->getData();
        return $pretty;
    }
    // }}}

    // {{{ toRaw()
    /**
     * Return the USMARC representation of the subfield
     *
     * @return string USMARC representation
     */
    function toRaw()
    {
        $result = FILE_MARC_SUBFIELD_INDICATOR.$this->code.$this->data;
        return $result;
    }
    // }}}

    // {{{ setCode()
    /**
     * Sets code of the subfield
     *
     * @param string $code new code for the subfield
     * @return string code 
     */
    function setCode($code)
    {
        if ($code) {
            // could check more stringently; m/[a-Z]/ or the likes
            $this->code = $code;
        } else {
            // code must be _something_; raise error
            return false;
        }
        return true;
    }
    // }}}

    // {{{ setData()
    /**
     * Sets data of the subfield
     *
     * @param string $data new data for the subfield
     * @return string data
     */
    function setData($data)
    {
        $this->data = $data;
        return true;
    }
    // }}}

    // {{{ isEmpty()
    /**
     * Checks whether the subfield is empty or not
     *
     * @return bool True or false
     */
    function isEmpty()
    {
        // There is data
        if ($this->data) {
            return false;
        }

        // There is no data
        return true;
    }
    // }}}

    // {{{ getErrors()
    /**
     * Get errors (if any) for this object 
     *
     * Returns the error stack.
     *
     * @return PEAR_ErrorStack returns the error stack
     */
    function getErrors()
    {
        return $GLOBALS["_File_MARC_stack"];
    }
    // }}}
}
// }}}

?>

